import configparser
import json
import os
import shutil
import subprocess
import sys
from pathlib import Path
from xml.sax.saxutils import escape

USER = 'git'

WWW_DIR = Path('/var/www/git/')
REPO_DIR = Path.home() / 'repos'
CONF_PATH = Path.home() / 'stagit.conf'


if os.environ.get('USER') != USER:
    sys.stderr.write('wrong user\n')
    sys.exit(1)


def create_sh_script(path, lines):
    if lines:
        with open(path, 'w') as fh:
            fh.write('#!/bin/sh\n')
            for line in lines:
                fh.write(line + '\n')
        path.chmod(0o775)
    elif path.exists():
        path.unlink()


def get_repo_dir(repo):
    return REPO_DIR / f'{repo}.git'


class Config:
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.read(CONF_PATH)

    def get(self, section, key):
        return self.config.get(section, key, fallback='')

    def getboolean(self, section, key):
        return self.config.getboolean(section, key, fallback=False)

    def can_ssh(self, repo, user):
        allowed = self.get(repo, 'ssh').split()
        return user in allowed or '@all' in allowed

    def iter_repos(self):
        for key in self.config:
            if key != 'DEFAULT':
                yield key


config = Config()


def update_repo(repo):
    repodir = get_repo_dir(repo)
    if not repodir.exists():
        repodir.mkdir(parents=True, exist_ok=True)
        subprocess.check_call(['git', 'init', '--bare'], cwd=repodir)

    export_ok = repodir / 'git-daemon-export-ok'
    post_update = []
    if config.getboolean(repo, 'http'):
        post_update.append(
            "python3 -c 'import os, stagit; "
            'stagit.render_repo(os.environ["STAGIT_REPO"])\''
        )
        export_ok.touch()
    else:
        if export_ok.exists():
            export_ok.unlink()

    if config.get(repo, 'post-update'):
        post_update.append(config.get(repo, 'post-update'))
    create_sh_script(repodir / 'hooks' / 'post-update', post_update)

    with open(repodir / 'description', 'w') as fh:
        fh.write(config.get(repo, 'desc'))
    with open(repodir / 'url', 'w') as fh:
        fh.write(config.get(repo, 'url'))


def render_index():
    with open(WWW_DIR / 'index.html', 'w') as fh:
        fh.write('<!DOCTYPE html>\n')
        fh.write('<html>\n<head>\n')
        fh.write('<meta charset="UTF-8">')
        fh.write('<meta name="viewport" content="width=device-width">\n')
        fh.write('<title>Repositories</title>\n')
        fh.write('<link rel="stylesheet" type="text/css" href="style.css" />\n')
        fh.write('</head>\n<body>\n')
        fh.write('<h1>Repositories</h1>')
        fh.write('<hr/>\n')
        fh.write('<main id="content">\n')
        fh.write('<table>\n<thead>\n')
        fh.write('<tr><th>Name</th><th>Description</th></tr>\n')
        fh.write('</thead>\n<tbody>\n')
        for repo in config.iter_repos():
            if config.getboolean(repo, 'http'):
                fh.write('<tr><td><a href="{}/">{}</a></td><td>{}</td></tr>\n'.format(
                    escape(repo), escape(repo), escape(config.get(repo, 'desc'))
                ))
        fh.write('</tbody>\n</table>\n')
        fh.write('</main>\n</body>\n</html>\n')


def render_json_index():
    with open(WWW_DIR / 'index.json', 'w') as fh:
        json.dump([
            {
                'name': repo,
                'description': config.get(repo, 'desc'),
            }
            for repo in config.iter_repos()
            if config.getboolean(repo, 'http')
        ], fh)


def render_repo(repo):
    print(f'Generating HTML for {repo}…')
    target_dir = WWW_DIR / repo
    if target_dir.is_dir():
        shutil.rmtree(target_dir)
    target_dir.mkdir(parents=True)
    subprocess.check_call(['stagit', get_repo_dir(repo)], cwd=target_dir)


def render_all():
    for repo in WWW_DIR.iterdir():
        path = WWW_DIR / repo
        if path.is_dir():
            shutil.rmtree(path)
    for repo in config.iter_repos():
        if config.getboolean(repo, 'http'):
            render_repo(repo)
    render_index()
    render_json_index()


if __name__ == '__main__':
    all_repos = set(config.iter_repos())
    phy_repos = set(p.stem for p in REPO_DIR.iterdir() if p.suffix == '.git')
    stale_repos = phy_repos - all_repos

    print('Updating repos …')
    for repo in all_repos:
        update_repo(repo)

    if stale_repos:
        print('Warning: stale files for deleted repos:', ', '.join(stale_repos))

    render_all()
